"""DeepPresenter - An agentic and reflective presentation generation system"""

__version__ = "1.1.16"
__author__ = "Hao Zheng"
__email__ = "wszh712811@gmail.com"

import os

assert os.name == "posix", "DeepPresenter only supports Linux and macOS"
